@@ -8,27 +8,26 @@
     :copyright: Copyright 2014 Mark Mikofski
     :license: BSD, see LICENSE for details.
 """
-from __future__ import unicode_literals, print_function
-from builtins import *
+from __future__ import unicode_literals
 from io import open  # for opening files with encoding in Python 2
 import os
 import re
 import sys
 from copy import copy
 from sphinx.util import logging
-
-# Pygments MatlabLexer is in pygments.lexers.math, but recommended way to load
-# is from lexers, which has LEXERS dictionary, which PyDev doesn't see.
-from pygments.lexers import MatlabLexer  # @UnresolvedImport
-# PyDev doesn't see Token methods and subtypes that are generated at runtime
+from sphinxcontrib.mat_lexer import MatlabLexer
 from pygments.token import Token
+from zipfile import ZipFile
+import xml.etree.ElementTree as ET
 
 logger = logging.getLogger('matlab-domain')
 
+modules = {}
+
 MAT_DOM = 'sphinxcontrib-matlabdomain'
 __all__ = ['MatObject', 'MatModule', 'MatFunction', 'MatClass',  \
            'MatProperty', 'MatMethod', 'MatScript', 'MatException', \
-           'MatModuleAnalyzer', 'MAT_DOM']
+           'MatModuleAnalyzer', 'MatApplication', 'MAT_DOM']
 
 # TODO: use `self.tokens.pop()` instead of idx += 1, see MatFunction
 
@@ -48,24 +47,19 @@ class MatObject(object):
     :param name: Name of MATLAB object.
     :type name: str
 
-    MATLAB objects can be :class:`MatModule`, :class:`MatFunction` or
-    :class:`MatClass`. :class:`MatModule` are just folders that define a psuedo
-    namespace for :class:`MatFunction` and :class:`MatClass` in that folder.
+    MATLAB objects can be :class:`MatModule`, :class:`MatFunction`,
+    :class:`MatApplication` or :class:`MatClass`.
+    :class:`MatModule` are just folders that define a psuedo
+    namespace for :class:`MatFunction`, :class:`MatApplication`
+    and :class:`MatClass` in that folder.
     :class:`MatFunction` and :class:`MatClass` must begin with either
     ``function`` or ``classdef`` keywords.
+    :class:`MatApplication` must be a ``.mlapp`` file.
     """
-
-    @staticmethod
-    def dbg(msg, *args):
-        """
-        Alternate for sphinx_dbg for test-mode.
-        """
-        print('\t<test-mode>\n' + msg % args)
-
     basedir = None
+    encoding = None
     sphinx_env = None
     sphinx_app = None
-    sphinx_dbg = dbg
 
     def __init__(self, name):
         #: name of MATLAB object
@@ -123,7 +117,7 @@ class MatObject(object):
         fullpath = os.path.join(MatObject.basedir, objname)  # objname fullpath
         # package folders imported over mfile with same name
         if os.path.isdir(fullpath):
-            mod = sys.modules.get(package)
+            mod = modules.get(package)
             if mod:
                 msg = '[%s] mod %s already loaded.'
                 logger.debug(msg, MAT_DOM, package)
@@ -136,11 +130,16 @@ class MatObject(object):
             mfile = fullpath + '.m'
             msg = '[%s] matlabify %s from\n\t%s.'
             logger.debug(msg, MAT_DOM, package, mfile)
-            return MatObject.parse_mfile(mfile, name, path)  # parse mfile
+            return MatObject.parse_mfile(mfile, name, path, MatObject.encoding)  # parse mfile
+        elif os.path.isfile(fullpath + '.mlapp'):
+            mlappfile = fullpath + '.mlapp'
+            msg = '[%s] matlabify %s from\n\t%s.'
+            logger.debug(msg, MAT_DOM, package, mlappfile)
+            return MatObject.parse_mlappfile(mlappfile, name, path)
         return None
 
     @staticmethod
-    def parse_mfile(mfile, name, path):
+    def parse_mfile(mfile, name, path, encoding=None):
         """
         Use Pygments to parse mfile to determine type: function or class.
 
@@ -150,45 +149,99 @@ class MatObject(object):
         :type name: str
         :param path: Path of module containing :class:`MatObject`.
         :type path: str
+        :param encoding: Encoding of the Matlab file to load (default = utf-8)
+        :type encoding: str
         :returns: :class:`MatObject` that represents the type of mfile.
 
         Assumes that the first token in the file is either one of the keywords:
         "classdef" or "function" otherwise it is assumed to be a script.
+
+        File encoding can be set using sphinx config ``matlab_src_encoding``
+        Default behaviour : replaces parsing errors with ? chars
         """
         # use Pygments to parse mfile to determine type: function/classdef
         # read mfile code
-        with open(mfile, 'r', encoding='utf-8') as code_f:
-            code = code_f.read().replace('\r\n', '\n')  # repl crlf with lf
+        if encoding is None:
+            encoding = 'utf-8'
+        with open(mfile, 'r', encoding=encoding, errors='replace') as code_f:
+            code = code_f.read().replace('\r\n', '\n')
 
-        # remove the top comment header (if there is one) from the code string
         full_code = code
+        # remove the top comment header (if there is one) from the code string
         code = MatObject._remove_comment_header(code)
-
-        # functions must be contained in one line, no ellipsis, classdef is OK
         code = MatObject._remove_line_continuations(code)
         code = MatObject._fix_function_signatures(code)
-        code = MatObject._fix_string_double_quotes(code)
 
-        # Produce tokes, but fix incorrect (Token.Keyword, 'function')
         tks = list(MatlabLexer().get_tokens(code))
-        tks = MatObject._fix_function_variables(tks)
 
         modname = path.replace(os.sep, '.')  # module name
+
         # assume that functions and classes always start with a keyword
-        if tks[0] == (Token.Keyword, 'function'):
-            logger.debug('[%s] parsing function %s from %s.', MAT_DOM,
-                                 name, modname)
-            return MatFunction(name, modname, tks)
-        elif tks[0] == (Token.Keyword, 'classdef'):
-            logger.debug('[%s] parsing classdef %s from %s.', MAT_DOM,
-                                 name, modname)
+        def isFunction(token):
+            return token == (Token.Keyword, 'function')
+
+        def isClass(token):
+            return token == (Token.Keyword, 'classdef')
+
+        if isClass(tks[0]):
+            logger.debug('[%s] parsing classdef %s from %s.', MAT_DOM, name, modname)
             return MatClass(name, modname, tks)
+        elif isFunction(tks[0]):
+            logger.debug('[%s] parsing function %s from %s.', MAT_DOM, name, modname)
+            return MatFunction(name, modname, tks)
         else:
             # it's a script file retoken with header comment
             tks = list(MatlabLexer().get_tokens(full_code))
             return MatScript(name, modname, tks)
         return None
 
+    @staticmethod
+    def parse_mlappfile(mlappfile, name, path):
+        """
+        Uses ZipFile to read the metadata/appMetadata.xml file and
+        the metadata/coreProperties.xml file description tags.
+        Parses XML content using ElementTree.
+
+        :param mlappfile: Full path of ``.mlapp`` file.
+        :type mlappfile: str
+        :param name: Name of :class:`MatApplication`.
+        :type name: str
+        :param path: Path of module containing :class:`MatApplication`.
+        :type path: str
+        :returns: :class:`MatApplication` representing the application.
+        """
+
+        # TODO: We could use this method to parse other matlab binaries
+
+        # Read contents of meta-data file
+        # This might change in different Matlab versions
+        with ZipFile(mlappfile, 'r') as mlapp:
+            meta = ET.fromstring(mlapp.read('metadata/appMetadata.xml'))
+            core = ET.fromstring(mlapp.read('metadata/coreProperties.xml'))
+
+        metaNs = { 'ns' : "http://schemas.mathworks.com/appDesigner/app/2017/appMetadata" }
+        coreNs = {
+                'cp': "http://schemas.openxmlformats.org/package/2006/metadata/core-properties",
+                'dc': "http://purl.org/dc/elements/1.1/",
+                'dcmitype': "http://purl.org/dc/dcmitype/",
+                'dcterms': "http://purl.org/dc/terms/",
+                'xsi': "http://www.w3.org/2001/XMLSchema-instance"
+                }
+
+        coreDesc = core.find('dc:description', coreNs)
+        metaDesc = meta.find('ns:description', metaNs)
+
+        doc = []
+        if coreDesc is not None:
+            doc.append(coreDesc.text)
+        if metaDesc is not None:
+            doc.append(metaDesc.text)
+        docstring = '\n\n'.join(doc)
+
+        modname = path.replace(os.sep, '.')  # module name
+
+        return MatApplication(name, modname, docstring)
+
     @staticmethod
     def _remove_comment_header(code):
         """
@@ -264,59 +317,6 @@ class MatObject(object):
         logger.debug(msg, MAT_DOM)
         return code
 
-    @staticmethod
-    def _fix_function_variables(tokens):
-        """ Fixes invalid `function` tokens
-
-        Pygments will mark any text matching `token` as Token.Keyword. This
-        function fixes this issue, by converting invalid marked functions as
-        Token.Text.
-
-        :param tokens:
-        :type code: list of Tokens
-        :return: Modified list of tokens
-        """
-        for idx, token in enumerate(tokens):
-            temp_token = (token[0], token[1].lstrip())
-            if temp_token == (Token.Keyword, 'function'):
-                previous_token = tokens[idx-1] if idx > 0 else None
-                if previous_token and not previous_token == (Token.Text, '\n'):
-                    # This is not a real function as it is not after a new line
-                    tokens[idx] = (Token.Text, token[1])
-                    continue
-                next_token = tokens[idx+1] if idx < len(tokens)-1 else None
-                if next_token and not re.match(r'[\s[]', next_token[1]):
-                    # This is not a real function as it is not followed by '\s' or '['
-                    tokens[idx] = (Token.Text, token[1])
-                    continue
-        return tokens
-
-    @staticmethod
-    def _fix_string_double_quotes(code):
-        """
-        Convert double quotes to single qoute strings, pygments doesn't know
-        these.
-
-        :param code:
-        :type code: str
-        :return: Code string where %s is removed from strings
-        """
-        pat = r"(?!\s*%)(.*)([\"].*[\"]{1})(.*)"
-        pat = re.compile(pat, re.X | re.MULTILINE)  # search start of every line
-
-        # replacement function
-        def repl(m):
-            retv = m.group(0)
-            if m.group(2):
-                retv = m.group(1) + "''" + m.group(3)
-            return retv
-
-        code = pat.sub(repl, code)  # search for string spec and apply replacement
-        return code
-
-
-
-
 
 # TODO: get docstring and __all__ from contents.m if exists
 class MatModule(MatObject):
@@ -339,7 +339,7 @@ class MatModule(MatObject):
         #: name of package (full path from basedir to module)
         self.package = package
         # add module to system dictionary
-        sys.modules[package] = self
+        modules[package] = self
 
     def safe_getmembers(self):
         results = []
@@ -955,10 +955,10 @@ class MatClass(MatMixin, MatObject):
                             self._tk_eq(idx, (Token.Punctuation, '=')) or
                             self._tk_eq(idx, (Token.Punctuation, '(')) or
                             self._tk_eq(idx, (Token.Punctuation, ')')) or
+                            self._tk_eq(idx, (Token.Punctuation, ';')) or
                             self._tk_eq(idx, (Token.Punctuation, ','))):
-                            msg = ['[%s] Skipping tokens for methods defined in separate files.',
-                                   'token #%d: %r']
-                            MatClass.sphinx_dbg('\n'.join(msg), MAT_DOM, idx, self.tokens[idx])
+                            msg = '[%s] Skipping tokens for methods defined in separate files.\ntoken #%d: %r'
+                            logger.debug(msg, MAT_DOM, idx, self.tokens[idx])
                             idx += 1 + self._whitespace(idx + 1)
                         elif self._tk_eq(idx, (Token.Keyword, 'end')):
                             idx += 1
@@ -1021,6 +1021,12 @@ class MatClass(MatMixin, MatObject):
                     idx += 1
 
                 idx += self._blanks(idx)  # skip blanks
+
+                # Continue if attribute is assigned a boolean value
+                if self.tokens[idx][0] == Token.Name.Builtin:
+                    idx += 1
+                    continue
+
                 # continue to next attribute separated by commas
                 if self._tk_eq(idx, (Token.Punctuation, ',')):
                     idx += 1
@@ -1111,11 +1117,14 @@ class MatClass(MatMixin, MatObject):
                 # search folders
                 for m in dirs:
                     # check if module has been matlabified already
-                    mod = sys.modules.get('.'.join([root_mod, m]).lstrip('.'))
+                    mod_name = '.'.join([root_mod, m]).lstrip('.')
+                    mod = modules.get(mod_name)
                     if not mod:
                         continue
                     # check if base class is attr of module
                     b_ = mod.getter(b, None)
+                    if not b_:
+                        b_ = mod.getter(b.lstrip(m.lstrip('+')), None)
                     if b_:
                         bases_[b] = b_
                         break
@@ -1244,6 +1253,35 @@ class MatScript(MatObject):
     def __module__(self):
         return self.module
 
+
+class MatApplication(MatObject):
+    """
+    Representation of the documentation in a Matlab Application.
+
+    :param name: Name of :class:`MatObject`.
+    :type name: str
+    :param modname: Name of folder containing :class:`MatObject`.
+    :type modname: str
+    :param desc: Summary and description string.
+    :type desc: str
+    """
+
+    def __init__(self, name, modname, desc):
+        super(MatApplication, self).__init__(name)
+        #: Path of folder containing :class:`MatApplication`.
+        self.module = modname
+        #: docstring
+        self.docstring = desc
+
+    @property
+    def __doc__(self):
+        return self.docstring
+
+    @property
+    def __module__(self):
+        return self.module
+
+
 class MatException(MatObject):
     def __init__(self, name, path, tks):
         super(MatException, self).__init__(name)
@@ -1283,7 +1321,7 @@ class MatModuleAnalyzer(object):
             if isinstance(entry, MatcodeError):
                 raise entry
             return entry
-        mod = sys.modules.get(modname)
+        mod = modules.get(modname)
         if mod:
             obj = cls.for_folder(mod.path, modname)
         else:
@@ -1320,7 +1358,7 @@ class MatModuleAnalyzer(object):
         attr_visitor_collected = {}
         attr_visitor_tagorder = {}
         tagnumber = 0
-        mod = sys.modules[self.modname]
+        mod = modules[self.modname]
         # walk package tree
         for k, v in mod.safe_getmembers():
             if hasattr(v, 'docstring'):
